VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "PathLinks"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'=============
'PathLinks 1.4
'=============
'
'Why PathLinks?
'==============
'
'People have been muddling along for years putting data files into App.Path
'and getting away with it fine much of the time.  But along came Windows
'Vista and the rules seemed to suddenly change!
'
'So now you can't safely put your read/write files into folders under
'App.Path anymore.  What to do?
'
'
'PathLinks To The Rescue!
'========================
'
'At program startup you make a call to PathLinks.Publish(), passing it the
'relative path of a folder or folders you install under App.Path when you
'install your program.
'
'The first run after installation should be elevated.  After that all users
'can reach your data subfolders without elevation.
'
'   * If running on a pre-Vista OS:
'
'       + The specified data files and subfolders are created under App.Path
'         if they do not already exist.
'
'       + A good result (plPublishEnum.plpContinue) is returned.
'
'   * If running on Vista or later for the first time:
'
'       + If running non-elevated an error result
'         (plPublishEnum.plpElevationRequired) is returned and you should ask
'         the user to run once as administrator.
'
'       + If running first time and elevated then PathLinks.Publish() will
'         try to:
'
'           o Locate the "Public" special folder.
'
'               NOTE: If this fails it stops and returns an error
'                     (plPublishEnum.plpFailed) which is fatal but should
'                     be extremely rare!
'
'           o Create a folder under "Public" named:
'
'                   - <App.CompanyName>\<App.ProductName>
'                   - <App.CompanyName>\<App.ExeName>
'               or:
'                   - <App.ExeName>
'
'             Depending on whether you have CompanyName and ProductName defined
'             in your project.
'
'           o Go through the list of one or more App.Path data files and
'             subfolders as passed to PathLinks.Publish() and for each one:
'
'                   - Move that that file or subfolder tree from App.Path to
'                     the new folder in "Public."
'
'                   - Create an empty subfolder under the new folder in
'                     "Public" if no existing data folder of that name was
'                     found under App.Path.
'
'                   - Create a symlink in App.Path named for the file or
'                     top-level name in the data subfolder path that points to
'                     the new location under the new folder in "Public."
'
'           o Return a good result (plPublishEnum.plpContinue).
'
'   * If running on Vista or later a subsequent time either elevated or not,
'     a good result (plPublishEnum.plpContinue) is returned.
'
'
'Misc. Notes
'===========
'
'   * This is not intended (or required) for a program that normally runs
'     elevated!
'
'   * Data your program only reads does not need the PathLinks treatment.
'
'   * Do Not use folder structures under App.Path that contain a "." because
'     we have to use "." in a name to decide whether an item is a file.  We
'     can't look at the item itself, since it might not be present after
'     the installer completes (i.e. a file created at application run time).
'
'   * By the same token, files should have a "." and an extension.
'
'   * For a file under a subfolder just use the subfolder name.
'
'   * Do not "publish" subfolders under App.Path that contain DLLs, OCXs, etc.
'     This is just for data files and folders you need read/write access to
'     for any user running your program.
'
'   * Create your setup package as you would on Windows 2000 or XP.  Include
'     any data files in the package to be deployed to a subfolder under
'     App.Path as always.
'
'   * The only extra step after installing is a user must run your program
'     elevated one time before anyone uses it normally.
'
'   * Once again, pre-Vista PathLinks does nothing except create empty data
'     subfolders for you as required.  Empty files are not created.  From
'     there your program runs as always.
'
'   * You could also write a separate small program that just calls
'     PathLinks.Publish() to be run elevated right after installing.  Then
'     your actual application program or programs will not need PathLinks in
'     them at all!
'
'
'Passing The Folder List
'=======================
'
'PathLinks.Publish() expects a Variant parameter.  This Variant should contain
'either a single String value or else an array of String values:
'
'   result = PathLinks.Publish("temp")
'
'   result = PathLinks.Publish(Array("a.txt", "temp", "dictionary\db"))
'
'
'What Does This Do For You?
'==========================
'
'If running pre-Vista, nothing.
'
'If running on Vista or later:
'
'   PathLinks will move your data files and subfolders into a program work
'   folder that it creates under the computer's "Public" folder.  This is a
'   read/write location for all logged-on users without elevation.
'
'   Then PathLinks will create a symlink for each file and subfolder in
'   App.Path, that points to the new location of your data subfolder.
'
'   Your program can use App.Path as always, and should not notice any
'   difference!
'

'---- Private Declarations -------------------------------------------

Private Const S_OK As Long = 0
Private Const SYMBOLIC_LINK_FLAG_DIRECTORY As Long = 1
Private Const FOLDERID_Public As String = "{DFDF76A2-C82A-4D63-906A-5644AC457385}"

Private Declare Function CLSIDFromString Lib "ole32" ( _
    ByVal lpszGuid As Long, _
    ByVal pGuid As Long) As Long

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" ( _
    ByVal Destination As Long, _
    ByVal Source As Long, _
    ByVal Length As Long)

Private Declare Sub CoTaskMemFree Lib "ole32" (ByVal hMem As Long)

Private Declare Function CreateSymbolicLink Lib "kernel32" Alias "CreateSymbolicLinkW" ( _
    ByVal lpSymlinkFileName As Long, _
    ByVal lpTargetFileName As Long, _
    ByVal dwFlags As Long) As Long

Private Declare Function GetVersion Lib "kernel32" () As Long

Private Declare Function IsUserAnAdmin Lib "shell32" () As Long

Private Declare Function lstrlen Lib "kernel32" Alias "lstrlenW" (ByVal ptr As Long) As Long

Private Declare Function SHGetKnownFolderPath Lib "shell32" ( _
    ByVal rfid As Long, _
    ByVal dwFlags As Long, _
    ByVal hToken As Long, _
    ByRef ppszPath As Long) As Long

Private Type GUID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(7) As Byte
End Type

Private Enum plStatusEnum
    plsOldOS
    plsWin6PlusElevated
    plsWin6PlusNotElevated
    plsFailed
End Enum

'---- Public Enums ---------------------------------------------------

Public Enum plPublishEnum
    plpContinue
    plpElevationRequired
    plpPublished
    plpFailed
End Enum

#If False Then 'IDE trick to preserve case of Enum values.
    Dim plpContinue, plpElevationRequired, plpFailed
#End If

'---- Private Data ---------------------------------------------------

Private mPfStatus As plStatusEnum
Private mStrPublicFolder As String

'---- Public Properties and Methods ----------------------------------

Public Function Publish(ByVal FilesFolders As Variant) As plPublishEnum
    Dim blnIsArray As Boolean
    Dim I As Integer
    Dim strCDCache As String
    
    blnIsArray = (VarType(FilesFolders) And vbArray) = vbArray
    If blnIsArray Then
        For I = LBound(FilesFolders) To UBound(FilesFolders)
            If VarType(FilesFolders(I)) <> vbString Then
                GoTo BadArgument
            ElseIf Left$(FilesFolders(I), 1) = "\" Then
                FilesFolders(I) = Mid$(FilesFolders(I), 2)
            End If
        Next
    Else
        If VarType(FilesFolders) <> vbString Then
            GoTo BadArgument
        ElseIf Left$(FilesFolders, 1) = "\" Then
            FilesFolders = Mid$(FilesFolders, 2)
        End If
    End If
            
    strCDCache = CurDir$()
    ChDir App.Path
    ChDrive App.Path
    
    Select Case mPfStatus
        Case plsOldOS
            If blnIsArray Then
                For I = LBound(FilesFolders) To UBound(FilesFolders)
                    If Not (Exists(FilesFolders(I)) Or MayBeFile(FilesFolders(I))) Then
                        CreateSubFolder App.Path, FilesFolders(I)
                    End If
                Next
            Else
                If Not (Exists(FilesFolders) Or MayBeFile(FilesFolders)) Then
                    CreateSubFolder App.Path, FilesFolders
                End If
            End If
            
            Publish = plpContinue
        
        Case plsWin6PlusNotElevated
            Publish = plpContinue 'Tentative return value.
            
            If blnIsArray Then
                For I = LBound(FilesFolders) To UBound(FilesFolders)
                    If Not Exists(mStrPublicFolder & "\" & FilesFolders(I)) Then
                        Publish = plpElevationRequired
                        Exit For
                    End If
                Next
            Else
                If Not Exists(mStrPublicFolder & "\" & FilesFolders) Then
                    Publish = plpElevationRequired
                End If
            End If
        
        Case plsWin6PlusElevated
            Publish = plpContinue 'Tentative return value.
            
            If blnIsArray Then
                For I = LBound(FilesFolders) To UBound(FilesFolders)
                    If CreateMoveLink(FilesFolders(I)) Then Publish = plpPublished
                Next
            Else
                If CreateMoveLink(FilesFolders) Then Publish = plpPublished
            End If
        
        Case plsFailed
            Publish = plpFailed
    End Select
            
    ChDir strCDCache
    ChDrive strCDCache
    
    Exit Function
    
BadArgument:
    Err.Raise 5, _
              "PathLinks.Publish", _
              "String or Variant Array of Strings required."
End Function

'---- Private Methods ------------------------------------------------

Private Function CreateMoveLink(ByVal FileFolder As String) As Boolean
    'Just in case we're running elevated and we've already done this,
    'avoid trying to repeat it!
    '
    'Return True if we move and link anything.
    
    If Not Exists(mStrPublicFolder & "\" & FileFolder) Then
        If Exists(FileFolder) Then
            If IsFile(FileFolder) Then
                MoveFile FileFolder
                CreateSymbolicLink StrPtr(FileFolder), _
                                   StrPtr(mStrPublicFolder & "\" & FileFolder), _
                                   0
            Else
                CreateSubFolder mStrPublicFolder, FileFolder
                MoveFiles FileFolder
                CreateSymbolicLink StrPtr(FileFolder), _
                                   StrPtr(mStrPublicFolder & "\" & FileFolder), _
                                   SYMBOLIC_LINK_FLAG_DIRECTORY
            End If
        Else
            If MayBeFile(FileFolder) Then
                CreateSymbolicLink StrPtr(FileFolder), _
                                   StrPtr(mStrPublicFolder & "\" & FileFolder), _
                                   0
            Else
                CreateSubFolder App.Path, FileFolder 'Might be a multilevel path, so we need path here.
                CreateSubFolder mStrPublicFolder, FileFolder
                MoveFiles FileFolder 'Nothing to move, but removes bottom-level folder.
                CreateSymbolicLink StrPtr(FileFolder), _
                                   StrPtr(mStrPublicFolder & "\" & FileFolder), _
                                   SYMBOLIC_LINK_FLAG_DIRECTORY
            End If
        End If
        CreateMoveLink = True
    End If
End Function

Private Sub CreateSubFolder(ByVal ParentPath As String, ByVal SubFolder As String)
    'Build SubFolder path (under ParentPath) if it doesn't already exist.
    Dim strBuiltPath As String
    Dim strNodes() As String
    Dim lngNode As Long
    
    strNodes = PathNodes(SubFolder)
    strBuiltPath = ParentPath
    For lngNode = 0 To UBound(strNodes)
        strBuiltPath = strBuiltPath & "\" & strNodes(lngNode)
        If Not Exists(strBuiltPath) Then
            MkDir strBuiltPath
        End If
    Next
End Sub

Private Function Exists(ByVal FileOrFolder) As Boolean
    On Error Resume Next
    GetAttr FileOrFolder
    Exists = Err.Number = 0
End Function

Private Function GetKnownFolder(ByVal KnownFolderID As String) As String
    Dim guidFOLDERID As GUID
    Dim pPath As Long
    Dim lngLen As Long

    If CLSIDFromString(StrPtr(KnownFolderID), VarPtr(guidFOLDERID)) = 0 Then
        If SHGetKnownFolderPath(VarPtr(guidFOLDERID), 0, 0, pPath) = 0 Then
            lngLen = lstrlen(pPath)
            GetKnownFolder = String$(lngLen, 0)
            CopyMemory StrPtr(GetKnownFolder), pPath, 2 * lngLen
            CoTaskMemFree pPath
            Exit Function
        End If
    End If
    
    GetKnownFolder = vbNullString
End Function

Private Function IsFile(ByVal PossibleFile As String) As Boolean
    'Use this when we can (item exists).
    IsFile = (GetAttr(PossibleFile) And vbDirectory) <> vbDirectory
End Function

Private Function MayBeFile(ByVal PossibleFile As String) As Boolean
    'We have to use this approach because the file or folder might not
    'have been created by the installer.
    MayBeFile = InStrRev(PossibleFile, ".") > 0
End Function

Private Sub MoveFile(ByVal File As String)
    FileCopy File, mStrPublicFolder & "\" & File
    Kill File
End Sub

Private Sub MoveFiles(ByVal Folder As String)
    Dim colSubFolders As Collection
    Dim strName As String
    Dim strFQName As String
    Dim varSubFolder As Variant
    
    Set colSubFolders = New Collection
    strName = Dir$(App.Path & "\" & Folder & "\*", vbNormal Or vbDirectory)
    
    Do Until Len(strName) = 0
        strFQName = App.Path & "\" & Folder & "\" & strName
        If (GetAttr(strFQName) And vbDirectory) = vbDirectory Then
            If Not (strName = "." Or strName = "..") Then
                colSubFolders.Add Folder & "\" & strName
            End If
        Else
            FileCopy strFQName, mStrPublicFolder & "\" & Folder & "\" & strName
            Kill strFQName
        End If
        strName = Dir$()
    Loop
        
    For Each varSubFolder In colSubFolders
        MkDir mStrPublicFolder & "\" & varSubFolder
        MoveFiles varSubFolder
    Next
    
    RmDir App.Path & "\" & Folder
End Sub

Private Function PathNodes(ByVal Path As String) As String()
    Dim strNodes() As String
    Dim lngNodeLast As Long
    
    strNodes = Split(Path, "\")
    lngNodeLast = UBound(strNodes)
    If lngNodeLast > -1 Then
        'Trim trailing empty node if any.
        If Len(strNodes(lngNodeLast)) = 0 Then
            ReDim Preserve strNodes(lngNodeLast - 1)
        End If
    End If
    PathNodes = strNodes
End Function

'---- Event Handlers -------------------------------------------------
    
Private Sub Class_Initialize()
    Dim strPublicBase As String
    Dim pPath As Long
    
    If (GetVersion() And &HFF&) >= 6 Then
        mStrPublicFolder = GetKnownFolder(FOLDERID_Public)
        If StrPtr(mStrPublicFolder) = 0 Then
            mPfStatus = plsFailed
        Else
            If Len(App.CompanyName) > 0 Then
                strPublicBase = App.CompanyName & "\"
            End If
            If Len(App.ProductName) > 0 Then
                strPublicBase = strPublicBase & App.ProductName
            Else
                strPublicBase = strPublicBase & App.EXEName
            End If
            If IsUserAnAdmin() Then
                mPfStatus = plsWin6PlusElevated
                CreateSubFolder mStrPublicFolder, strPublicBase
            Else
                mPfStatus = plsWin6PlusNotElevated
            End If
            mStrPublicFolder = mStrPublicFolder & "\" & strPublicBase
        End If
    Else
        mPfStatus = plsOldOS
    End If
End Sub
